/*
 [The "BSD licence"]
 Copyright (c) 2005-2006 Terence Parr
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.antlr_yggdrasil.tool;

import antlr.Token;

import org.antlr_yggdrasil.analysis.Label;

import java.util.Iterator;
import java.util.List;
import java.util.Set;

public class NameSpaceChecker {
	protected Grammar grammar;

	public NameSpaceChecker(Grammar grammar) {
		this.grammar = grammar;
	}

	public void checkConflicts() {
		for (int i = 0; i < grammar.ruleIndexToRuleList.size(); i++) {
			String ruleName = (String) grammar.ruleIndexToRuleList.elementAt(i);
			if ( ruleName==null ) {
				continue;
			}
			Rule r = grammar.getRule(ruleName);
			checkForRuleDefinitionProblems(r);
		}
		// check all global scopes against tokens
		Iterator it = grammar.getGlobalScopes().values().iterator();
		while (it.hasNext()) {
			AttributeScope scope = (AttributeScope) it.next();
			checkForGlobalScopeTokenConflict(scope);
		}
		// check for missing rule, tokens
		lookForReferencesToUndefinedSymbols();
	}

	protected void checkForRuleDefinitionProblems(Rule r) {
		String ruleName = r.name;
		antlr.Token ruleToken = r.tree.getToken();
		int msgID = 0;
		if ( grammar.type==Grammar.PARSER && Character.isUpperCase(ruleName.charAt(0)) ) {
			msgID = ErrorManager.MSG_LEXER_RULES_NOT_ALLOWED;
        }
        else if ( grammar.type==Grammar.LEXER &&
			      Character.isLowerCase(ruleName.charAt(0)) &&
			      !r.isSynPred )
		{
			msgID = ErrorManager.MSG_PARSER_RULES_NOT_ALLOWED;
        }
		else if ( grammar.getGlobalScope(ruleName)!=null ) {
			msgID = ErrorManager.MSG_SYMBOL_CONFLICTS_WITH_GLOBAL_SCOPE;
		}
		if ( msgID!=0 ) {
			ErrorManager.grammarError(msgID, grammar, ruleToken, ruleName);
		}
	}

	/** If ref to undefined rule, give error at first occurrence.
	 *
	 *  If you ref ID in a combined grammar and don't define ID as a lexer rule
	 *  it is an error.
	 */
	protected void lookForReferencesToUndefinedSymbols() {
		// for each rule ref, ask if there is a rule definition
		for (Iterator iter = grammar.ruleRefs.iterator(); iter.hasNext();) {
			Token tok = (Token) iter.next();
			String ruleName = tok.getText();
			if ( grammar.getRule(ruleName)==null &&
			     grammar.getTokenType(ruleName)!=Label.EOF )
			{
				ErrorManager.grammarError(ErrorManager.MSG_UNDEFINED_RULE_REF,
										  grammar,
										  tok,
										  ruleName);
			}
        }
		if ( grammar.type==Grammar.COMBINED ) {
			for (Iterator iter = grammar.tokenIDRefs.iterator(); iter.hasNext();) {
				Token tok = (Token) iter.next();
				String tokenID = tok.getText();
				if ( !grammar.lexerRules.contains(tokenID) &&
					 grammar.getTokenType(tokenID)!=Label.EOF )
				{
					ErrorManager.grammarError(ErrorManager.MSG_NO_TOKEN_DEFINITION,
											  grammar,
											  tok,
											  tokenID);
				}
			}
		}
	}

	protected void checkForGlobalScopeTokenConflict(AttributeScope scope) {
		if ( grammar.getTokenType(scope.getName())!=Label.INVALID ) {
			ErrorManager.grammarError(ErrorManager.MSG_SYMBOL_CONFLICTS_WITH_GLOBAL_SCOPE,
									  grammar, null, scope.getName());
		}
	}

}
